from loopy import Reduction

from dune.codegen.generation import (basis_mixin,
                                     kernel_cached,
                                     instruction,
                                     temporary_variable,
                                     globalarg,
                                     class_member,
                                     initializer_list,
                                     include_file,)
from dune.codegen.tools import get_pymbolic_basename, get_pymbolic_indices
from dune.codegen.loopy.target import type_floatingpoint
from dune.codegen.pdelab.basis import (GenericBasisMixin,
                                       declare_cache_temporary,
                                       name_localbasis_cache,
                                       type_localbasis,
                                       FEM_name_mangling)
from dune.codegen.pdelab.driver import (isPk,
                                        isQk,
                                        )
from dune.codegen.pdelab.geometry import world_dimension, component_iname
from dune.codegen.pdelab.spaces import type_leaf_gfs, name_lfs
from dune.codegen.pdelab.restriction import restricted_name
from dune.codegen.blockstructured.spaces import lfs_inames
from dune.codegen.blockstructured.tools import tensor_index_to_sequential_index, sub_element_inames

from ufl import MixedElement

import pymbolic.primitives as prim


@basis_mixin("blockstructured")
class BlockStructuredBasisMixin(GenericBasisMixin):
    def lfs_inames(self, element, restriction, number, context=""):
        return lfs_inames(element, restriction, number, context=context)

    def implement_basis(self, element, restriction, number, context=''):
        assert not isinstance(element, MixedElement)
        name = "phi_{}".format(FEM_name_mangling(element))
        name = restricted_name(name, restriction)
        self.evaluate_basis(element, name, restriction)
        inames = self.lfs_inames(element, restriction, number, context=context)

        return prim.Subscript(prim.Variable(name), (tensor_index_to_sequential_index(inames, element.degree() + 1), 0))

    @kernel_cached
    def evaluate_basis(self, element, name, restriction):
        temporary_variable(name, shape=((element.degree() + 1)**world_dimension(), 1),
                           decl_method=declare_cache_temporary(element, restriction, 'Function'))
        cache = name_localbasis_cache(element)
        qp = self.to_cell(self.quadrature_position_in_micro())
        localbasis = name_localbasis(element)
        instruction(inames=self.quadrature_inames(),
                    code='{} = {}.evaluateFunction({}, {});'.format(name, cache, str(qp), localbasis),
                    assignees=frozenset({name}),
                    read_variables=frozenset({get_pymbolic_basename(qp)}),
                    )

    def implement_reference_gradient(self, element, restriction, number, context=''):
        assert not isinstance(element, MixedElement)
        name = "js_{}".format(FEM_name_mangling(element))
        name = restricted_name(name, restriction)
        self.evaluate_reference_gradient(element, name, restriction)
        inames = self.lfs_inames(element, restriction, number, context=context)

        return prim.Subscript(prim.Variable(name), (tensor_index_to_sequential_index(inames, element.degree() + 1), 0))

    @kernel_cached
    def evaluate_reference_gradient(self, element, name, restriction):
        temporary_variable(name, shape=((element.degree() + 1)**world_dimension(), 1, world_dimension()),
                           decl_method=declare_cache_temporary(element, restriction, 'Jacobian'))
        cache = name_localbasis_cache(element)
        qp = self.to_cell(self.quadrature_position_in_micro())
        localbasis = name_localbasis(element)
        instruction(inames=self.quadrature_inames(),
                    code='{} = {}.evaluateJacobian({}, {});'.format(name, cache, str(qp), localbasis),
                    assignees=frozenset({name}),
                    read_variables=frozenset({get_pymbolic_basename(qp)}),
                    )

    @kernel_cached
    def evaluate_coefficient_gradient(self, element, name, container, restriction, index):
        sub_element = element
        if isinstance(element, MixedElement):
            sub_element = element.extract_component(index)[1]
        from ufl import FiniteElement
        assert isinstance(sub_element, FiniteElement)

        temporary_variable(name, shape=(element.cell().geometric_dimension(),), managed=True)

        dimindex = component_iname(count=0)

        lfs = name_lfs(element, restriction, index)
        basis = self.implement_reference_gradient(sub_element, restriction, 0, context='trialgrad')
        basisindex = get_pymbolic_indices(basis)[:-1]
        from dune.codegen.tools import maybe_wrap_subscript
        basis = maybe_wrap_subscript(basis, prim.Variable(dimindex))

        from dune.codegen.blockstructured.argument import pymbolic_coefficient
        coeff = pymbolic_coefficient(container, lfs, sub_element, basisindex)

        assignee = prim.Subscript(prim.Variable(name), (prim.Variable(dimindex),))
        reduction_expr = prim.Product((coeff, basis))

        instruction(expression=Reduction("sum", basisindex, reduction_expr, allow_simultaneous=True),
                    assignee=assignee,
                    forced_iname_deps=frozenset(self.quadrature_inames() + (dimindex,) + sub_element_inames()),
                    forced_iname_deps_is_final=True,
                    )

    @kernel_cached
    def evaluate_coefficient(self, element, name, container, restriction, index):
        sub_element = element
        if isinstance(element, MixedElement):
            sub_element = element.extract_component(index)[1]

        from ufl import FiniteElement
        assert isinstance(sub_element, FiniteElement)

        temporary_variable(name, shape=(), managed=True)

        lfs = name_lfs(element, restriction, index)
        basis = self.implement_basis(sub_element, restriction, 0, context='trial')
        basisindex = get_pymbolic_indices(basis)[:-1]

        from dune.codegen.blockstructured.argument import pymbolic_coefficient
        coeff = pymbolic_coefficient(container, lfs, sub_element, basisindex)

        assignee = prim.Variable(name)
        reduction_expr = prim.Product((coeff, basis))

        instruction(expression=Reduction("sum", basisindex, reduction_expr, allow_simultaneous=True),
                    assignee=assignee,
                    forced_iname_deps=frozenset(self.quadrature_inames() + sub_element_inames()),
                    forced_iname_deps_is_final=True,
                    )


# define FE basis explicitly in localoperator
@class_member(classtag="operator")
def typedef_localbasis(element, name):
    df = "typename {}::Traits::GridView::ctype".format(type_leaf_gfs(element))
    r = type_floatingpoint()
    dim = world_dimension()
    if isPk(element):
        if dim == 1:
            include_file("dune/localfunctions/lagrange/pk1d/pk1dlocalbasis.hh", filetag="operatorfile")
            basis_type = "Pk1DLocalBasis<{}, {}, {}>".format(df, r, element._degree)
        elif dim == 2:
            include_file("dune/localfunctions/lagrange/pk2d/pk2dlocalbasis.hh", filetag="operatorfile")
            basis_type = "Pk2DLocalBasis<{}, {}, {}>".format(df, r, element._degree)
        elif dim == 3:
            include_file("dune/localfunctions/lagrange/pk3d/pk3dlocalbasis.hh", filetag="operatorfile")
            basis_type = "Pk3DLocalBasis<{}, {}, {}>".format(df, r, element._degree)
        else:
            raise NotImplementedError("P{} in {}D is not implemented".format(element._degree, dim))
    elif isQk(element):
        include_file("dune/localfunctions/lagrange/qk/qklocalbasis.hh", filetag="operatorfile")
        basis_type = "QkLocalBasis<{}, {}, {}, {}>".format(df, r, element._degree, dim)
    else:
        raise NotImplementedError("Element type not known in code generation")
    return "using {} = Dune::{};".format(name, basis_type)


@class_member(classtag="operator")
def define_localbasis(leaf_element, name):
    localBasis_type = type_localbasis(leaf_element)
    initializer_list(name, (), classtag="operator")
    return "const {} {};".format(localBasis_type, name)


def name_localbasis(leaf_element):
    name = "{}_microElementBasis".format(FEM_name_mangling(leaf_element))
    globalarg(name)
    define_localbasis(leaf_element, name)
    return name
